/* PersistentCacheVector.java */
package jwiley.effective;

import java.io.*;
import java.util.*;

/**
 *  A replacement for a Vector that uses an LRU cache and persists all
 *  entries greater than the cache size.
 */
public class PersistentCacheVector implements Cloneable, Serializable
{
   /** count of the number of these objects created. */
   static long lCount;
   /** Maximum size the cache is allowed to grow to. */
   int iMaxCacheSize;
   /** Current cache size. */
   int iCacheSize;
   /** The cache. */
   Hashtable cache = new Hashtable();
   /** The Vector mirror. This will only store a 
       VectorEntry object for each entry inserted. */
   Vector rows = new Vector();
   /** ObjectFile that stores the persisted objects.
       @see ObjectFile
   */
   ObjectFile of;
   /** Name of file where persistent objects are stored. */
   String sTmpName;
   /** First entry in the Least Recently Used (LRU) linked list.  The first
    entry represents the most recently accessed Cache element. */
   CacheEntry firstCacheEntry;
   /** Last entry in the Least Recently Used (LRU) linked list. The last
    entry represents the least recently accessed cache element. */
   CacheEntry lastCacheEntry;

   /**
    * An inner class that represents a minimal entry in the Array.
    */
   class StubEntry
   {
      /** flag that indicates if entry is in the cache or on disk. */
      boolean inCache;
      /** If entry on disk, this is where it was stored. */
      long filePointer = -1;
   }

   /**
    * An inner class that represents an entry in the cache.
    */
   class CacheEntry
   {
      /** Key that maps to the element. We use an Integer object
          that corresponds to the index in the Array. */
      Integer key;
      /** Object to store in cache. */
      Object o;
      /** References that support a doubly linked list. */
      CacheEntry prev, next;
   }

   /**
    * Constructor that sets the cache size.  File is not opened unless
    * array grows greater than cache size.
    * @param iCacheSize Maximum size the cache should be allowed to grow to.
    */
   public PersistentCacheVector(int iCacheSize)
   {
      this.iMaxCacheSize = iCacheSize;
   }

   private final void openTempFile() throws IOException
   {
      boolean bInValid = true;

      while (bInValid)
      {
         sTmpName = "tmp" + (System.currentTimeMillis()) 
                          + (++lCount) + ".obf";
         File f = new File(sTmpName);
         if (!f.exists())
            bInValid = false;
      }

      of = new ObjectFile(sTmpName);
   }

   /**
    * Method to add an object to this Vector.  We add the
    * object to either the cache or to disk.
    * @param o The object to add to the array.
    */
   public final void add(Serializable o) throws IOException
   {
      StubEntry e = new StubEntry();

      rows.add(e);

      if (iCacheSize < iMaxCacheSize)
      {
         e.inCache = true;
         CacheEntry ce = new CacheEntry();
         ce.o = o;
         ce.key = new Integer(rows.size() - 1);
         cache.put(ce.key, ce);
         iCacheSize++;

         // add Cache Entry to LRU list
         if (lastCacheEntry != null)
         {
            // add to front of list
            ce.next = firstCacheEntry;
            firstCacheEntry.prev = ce;
            firstCacheEntry = ce;
         }
         else
         {
            // empty list
            firstCacheEntry = lastCacheEntry = ce;
         }
      }
      else
      {
        if (of == null)
            openTempFile();

        e.filePointer = of.writeObject(o);
      }
   }

   /**
    * Vector method to access an element at a specific index.
    * We retrieve the object from either the cache or from the object file.
    * @param idx The array index of the element to retrieve.
    * @returns The Object originally stored.
    */
   public final Object get(int idx) throws IndexOutOfBoundsException,
                                           IOException
   {
      if (idx < 0 || idx >= rows.size())
         throw new IndexOutOfBoundsException("Index: " + idx 
                                         + " out of bounds.");

      StubEntry e = (StubEntry) rows.get(idx);

      Object o=null;

      if (e.inCache)
      {
         // get it
         CacheEntry ce = null;
         ce = (CacheEntry) cache.get(new Integer(idx));

         if (ce == null)
            throw new IOException("Element at idx " + idx + " is NULL!");

         if ((ce != null && ce.o == null))
            throw new IOException("Cache Element's object at idx " 
                                  + idx + " NOT in cache!");

         o = ce.o;

         if (ce != firstCacheEntry)
         {
            // remove it from its current place in list
            if (ce.next != null)
               ce.next.prev = ce.prev;
            else // this is the last one in the cache
                lastCacheEntry = ce.prev;

            ce.prev.next = ce.next;

            // move it to front of list
            ce.next = firstCacheEntry;
            ce.prev = null;
            firstCacheEntry.prev = ce;
            firstCacheEntry = ce;
         }
      }
      else
      {
         // this will now be in the cache
         e.inCache = true;

         // retrieve and put in cache
         try
         {
            o = of.readObject(e.filePointer);
         } catch (ClassNotFoundException cnfe)
           { throw new IOException(cnfe.getMessage()); }

         //*** Check if cache is FULL!
         if (iCacheSize == iMaxCacheSize)
         {
            // REMOVE LRU (tail of list).
            CacheEntry leastUsed = lastCacheEntry;
            if (leastUsed.prev != null)
            {
               leastUsed.prev.next = null;
               lastCacheEntry = leastUsed.prev;
               lastCacheEntry.next = null;
            }
            else
            {
               // removing the only entry
               firstCacheEntry = lastCacheEntry = null;
            }

            // add retrieved object to cache
            CacheEntry ce = new CacheEntry();
            ce.o = o;
            ce.key = new Integer(idx);
            cache.put(ce.key, ce);

            // add Cache Entry to LRU list
            if (lastCacheEntry != null)
            {
               // add to front of list
               ce.next = firstCacheEntry;
               firstCacheEntry.prev = ce;
               firstCacheEntry = ce;
            }
            else
            {
               // empty list
               firstCacheEntry = lastCacheEntry = ce;
            }

            // get StubEntry for one to remove
            StubEntry outStubEntry = (StubEntry) 
                rows.get(leastUsed.key.intValue());

            // get Object out of cache
            CacheEntry outCacheEntry = (CacheEntry) 
                cache.remove(leastUsed.key);
            if (outCacheEntry == null)
               throw new RuntimeException("Cache Entry at " 
                              + leastUsed.key + " is Null!");

            if (outCacheEntry != null && outCacheEntry.o == null)
               throw new RuntimeException("Cache object at " 
                               + leastUsed.key + " is Null!");

            Object outObject = outCacheEntry.o;

            outStubEntry.inCache = false;

            if (outStubEntry.filePointer == -1)
            {
               // had gone right to cache
                  outStubEntry.filePointer = 
                           of.writeObject((Serializable)outObject);
            }
            else
            {
               // already in the file - size changed?
                int iCurrentSize = 
                  of.getObjectLength(outStubEntry.filePointer);

                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                ObjectOutputStream oos = new ObjectOutputStream(baos);
                oos.writeObject((Serializable) outObject);
                oos.flush();
                int datalen = baos.size();

                if (datalen <= iCurrentSize)
                    of.rewriteObject(outStubEntry.filePointer, 
                                     baos.toByteArray());
                else
                    outStubEntry.filePointer = 
                                of.writeObject((Serializable)outObject);

                baos = null;
                oos = null;
                outObject = null;
            }
         }
         else
         {
            CacheEntry ce = new CacheEntry();
            ce.o = o;
            ce.key = new Integer(idx);
            cache.put(ce.key, ce);
            iCacheSize++;

            // add Cache Entry to LRU list
            if (lastCacheEntry != null)
            {
               // add to front of list
               ce.next = firstCacheEntry;
               firstCacheEntry.prev = ce;
               firstCacheEntry = ce;
            }
            else
            {
               // empty list
               firstCacheEntry = lastCacheEntry = ce;
            }
         }

      }

      return o;
   }

   /**
    *  Remove an Object in the vector at the specified index.
    */
   public final Object remove(int index) 
                              throws IndexOutOfBoundsException,
                                     IOException
   {
      Object removedObject = null;

      StubEntry e = (StubEntry) rows.remove(index);

      if (e == null)
         throw new IOException("Entry is null at index: " + index);

      if (e.inCache)
      {
         CacheEntry ce = (CacheEntry) cache.remove(new Integer(index));

         if (ce != null)
         {
            removedObject = ce.o;

            // remove from linked list
            if (ce.next != null)
               ce.next.prev = ce.prev;
            else
            {
               // removing last on list.
               lastCacheEntry = ce.prev;
            }

            if (ce.prev != null)
               ce.prev.next = ce.next;
            if (firstCacheEntry == ce)
               firstCacheEntry = ce.next;
            ce = null;
            // Cache now has an opening
            iCacheSize--;
         }
      }
      else
      {
        try
        {
            removedObject = of.readObject(e.filePointer);
        } catch (ClassNotFoundException cnfe)
          { throw new IOException(cnfe.getMessage()); }
         e.filePointer = -1;
      }

      // do a decrement of all elements after index + 1
      modifyCacheElements(index + 1,1,false, true);

      return removedObject;
   }

   // for debugging
   private void printCacheList(boolean forward)
   {
      System.out.println("Printing list " + 
                 (forward?"forward":"backwards"));

      CacheEntry tmp= forward?firstCacheEntry: lastCacheEntry;
      while (tmp != null)
      {
         System.out.print(tmp.key + ",");
         tmp = forward?tmp.next:tmp.prev;
      }
      System.out.println("");
   }

   // inclusive modify
   private final void modifyCacheElements(int index, int quantity, 
                                  boolean increment, boolean after)
   {
      CacheEntry traverse = firstCacheEntry;

      while (traverse != null)
      {
         int curValue = traverse.key.intValue();
         if ((after?curValue >= index : curValue <= index))
         {
            // remove old
            cache.remove(traverse.key);

            // now modify key
            traverse.key = new Integer(increment?curValue 
                               + quantity:curValue - quantity);
         }
         traverse = traverse.next;
      }

      // go through a second time, adding new keys
      traverse = firstCacheEntry;
      while (traverse != null)
      {
         // reverse operation
         int curValue = traverse.key.intValue();
         int oldValue = increment?curValue - quantity:curValue + quantity;

         if ((after?oldValue >= index : oldValue <= index))
         {
            // add new
            cache.put(traverse.key, traverse);
         }
         traverse = traverse.next;
      }
   }

   /**
    *  Copy entire contents of this Vector into 
    */
   public final void copyTo(Object oArray[]) throws IOException
   {
      if (oArray.length < rows.size())
         return;

      for (int i=0; i < oArray.length; i++)
         oArray[i] = this.get(i);
   }

   /**
    *  Close and delete potentially large object file.
    */
   public final void close()
   {
      try
      {
         of.close();
         File f = new File(sTmpName);
         if (f.exists())
            f.delete();
      } catch (Exception e) { }
   }

   /**
    *  On finalization close().
    */
   public void finalize()
   {
      try
      {
         super.finalize();
         close();
      } catch (Throwable t) { }
   }

   /**
    *  Return the size of the Vector.
    */
   public final int size()
   {
      return rows.size();
   }

   /**
    *  Main method.
    */
   public static void main(String args[])
   {
      try
      {
         PersistentCacheVector pcv = new PersistentCacheVector(200);

         System.out.println("adding 300 objects...");
         // test add
         for (int i = 0; i < 300; i++)
            pcv.add(new Integer(i));
         System.out.println("Size: " + pcv.size());

         // test get
         System.out.println("Testing get...");
         for (int i = 0; i < 6; i++)
            System.out.print(pcv.get(i) + ",");
         System.out.println();

         // test a remove
         pcv.remove(0);

         // now get one
         System.out.println("Now the element at index 5 is : " 
                            + pcv.get(5));
         System.out.println("Size is: " + pcv.size());

         // now get some at the end
         for (int i=295; i < pcv.size(); i++)
            System.out.print(pcv.get(i) + ",");
         System.out.println();

         // now get some at the beginning
         for (int i=0; i < 5; i++)
            System.out.print(pcv.get(i) + ",");

         // test remove
         System.out.println("Testing remove...");
         System.out.println("Size: " + pcv.size());
         System.out.println("Removing 10");
         for (int i = 0; i < 10; i++)
            pcv.remove(i);
         System.out.println("Size: " + pcv.size());

         // examine first 10
         System.out.println("First 10.");
         for (int i=0; i < 10; i++)
            System.out.print(pcv.get(i) + ",");
         System.out.println();
         
         pcv.close();
      } catch (Exception e)
        {
           e.printStackTrace();
        }
   }
}

